# 负载均衡算法

本文介绍各种常见的负载均衡算法，说明它们的优缺点、适用场景，并给出相应的实现代码。

接着文末会探讨下负载不均衡的常见原因，以及解决方法。

## 一、 常见负载均衡算法

## 1. 加权轮询 - Weighted Round Robin

适用于需要进行流量切分的场景。

### 1.1 基于最大公约数的加权轮询算法

> http://kb.linuxvirtualserver.org/wiki/Weighted_Round-Robin_Scheduling

此算法在某些特殊情况下负载不均衡，LVS 以及 APISIX 都使用此算法。

相关代码：

- [lua-resty-balancer - roundrobin.lua](https://github.com/openresty/lua-resty-balancer/blob/master/lib/resty/roundrobin.lua):
  APISIX 使用了这份实现。

### 1.2 平滑加权轮询算法 - smooth weighted round-robin

假设有 n 台服务器，权重分别为，且总权重为 100：

- m1 = w1
- m2 = w2
- ...
- m(n) = w(n)

那么我们有：

- w1 + w2 + ... + w(n) = 100

平滑加权轮询算法的策略为，每次请求时，首先

相关代码：

- [nginx - Upstream: smooth weighted round-robin balancing](https://github.com/phusion/nginx/commit/27e94984486058d73157038f7950a0a36ecc6e35):
  Nginx 的 SWRR 实现

## 2. 最小连接数 - Least Connections

加权轮询并不考虑 upstreams 的负载，可能会导致部分 upstream 负载过高（overburdened）。而「最小连接
数」用连接数来评判 upstream 的负载情况，可以使 upstreams 的负载更均衡。

理论上最小连接数在任何情况下都可以用于替换加权轮询，在请求处理时间长短不一的场景下尤其如此。

据 Istio 与 Enovy 团队试验发现，**基本在所有场景下，「最小连接数」的表现都要优于「加权轮询**」！因此
Istio 在 1.14 将默认算法从 ROUND_ROBIN 改成了 LEAST_REQUEST.

相关代码：

- [apisix - least_conn.lua](https://github.com/apache/apisix/blob/master/apisix/balancer/least_conn.lua):
  APISIX 的 least_conn 实现

## 3. 一致性哈希 - Consistent Hashing

一致性哈希算法在1997 年由麻省理工学院提出，是一种特殊的哈希算法。它有这几个特点：

- Session 亲和：在节点数不变的情况下，带有相同 key 的 requests 总会被映射到同一台 server。
- 最小改变：它在移除或者添加一个服务器时，能够尽可能小地改变已存在的 requests 与 servers 之间的映射
  关系。

因此它**非常适合重度依赖本地缓存的服务**。

相关文档：

- [一致性哈希 （Consistent Hashing）的前世今生](https://candicexiao.com/consistenthashing/)
- [Web caching with consistent hashing - 论文](https://www.ra.ethz.ch/cdstore/www8/data/2181/pdf/pd1.pdf)

相关代码：

- [lua-resty-balancer - chash.lua](https://github.com/openresty/lua-resty-balancer/blob/master/lib/resty/chash.lua):
  APISIX 使用了这份实现
- [Envoy - upstream/ring_hash_lb.c](https://github.com/envoyproxy/envoy/blob/v1.20.1/source/common/upstream/ring_hash_lb.cc):
  Envoy 的 ketama 实现
  - 完整的 PR: [HTTP consistent hash routing](https://github.com/envoyproxy/envoy/pull/496)

### 3.1 环割法 - Ring Hash / Ketama

Nginx/APISIX/Envoy 都是使用的 ketama 算法实现的一致性哈希，此算法的特点如下：

- 优点：
  - 实现简单，被广泛使用
  - 复杂度：算法复杂度为 `log(v * n)`，其中 n 为节点数，n 为每个节点的虚拟节点数
- 缺点：
  - 在虚拟节点数较少时，扩缩容的平衡性较差。被缩容节点的原有请求量，不能很均衡地分配到剩余的节点上。
  - Nginx/APISIX 硬编码了常量 `160` 作为虚拟节点数
  - Envoy 则要求用户通过 `minimum_ring_size` 以及 `maximum_ring_size` 自行定义此参数，并通过监控查看
    实际的虚拟节点数，确认达成最佳效果。

### 3.2 Maglev 算法

这是 Google 公开的一种一致性 Hash 算法，Envoy 可以选择使用此算法。

代码：

- [Envoy - upstream/maglev_lb.c](https://github.com/envoyproxy/envoy/blob/v1.20.1/source/common/upstream/maglev_lb.cc)

## 4. ewma 算法

ewma 是 twitter 使用的一种负载均衡算法，它会使用每个 upstream 最后一次请求的响应时间来为其生成权重。

它的基本思路是，响应得更快的 upstreams 负载会更低，应该把请求更多的派发给这类 upstreams.

如果你的服务使用的是 LEAST_CONN/ROUND_ROBIN 算法，发现请求速率很均衡，但是负载却不怎么均衡，或许可以
考虑下这个算法。

代码：

- [APISIX - ewma](https://github.com/apache/apisix/blob/master/apisix/balancer/ewma.lua)

## 负载不均衡问题及其排查方法

- 算法问题
- 机器性能不一致
  - CPU/MEM/Disk 的性能不一致，比如型号不一，或者老化
  - 机器负载不一致，其他租户在抢占资源
- 延迟问题
- QPS 规律性地波动，根据算法的不同，有可能会导致负载不均衡？

## 参考

- [深入浅出负载均衡 - vivo 互联网技术](https://mp.weixin.qq.com/s/RFjTDNsGiqkMNnkjVus5Zw)
