## 各语言通用的一些知识点

### 无符号数

Go/Rust/C/C++ 支持无符号数，但是无符号数非常危险，很容易搞出预期外的问题。因此 Java/Python 等语言不
提供对无符号数的支持。

只应该在两种场景下使用无符号数：

- 需要执行按位运算时
  - 比如移位 `1 << 10`、按位与 `0xa & 0xb`、按位或 `0xa | 0xb`、按位异或 `0xa ^ 0xb`
- 需要执行一些不寻常的算术操作时

因此无符号数一般只被用在解析二进制数据、密码学等场景下。

如果你觉得自己只需要用到非负数，那你通常应该使用 `int` 而不是无符号数！

### 浮点数

浮点数的规范为 IEEE 754，一般常见的是两种浮点数：

- float32: 单精度
- float64: 双精度

选用规则：除非你很确定 float32 就够用了，否则请使用 float64

浮点数的特点有：

- 存在精度误差，IEEE754 跟 10 进制的浮点数之间无法精确对应
  - 因此绝对不要直接在浮点数上使用 `==` 进行比较！
- 存在几个特殊值 `0` `-0` `NaN` `+Inf` `-Inf`

### Unicode 字符处理

目前最通用的字符集 UTF-8 是变长字符集，一个字符可能由 1-4 个字节表示，需要使用 utf-8 专用的方法来处
理 utf-8 数据。

python3 的字符串默认就会帮你处理 utf-8 解析相关的逻辑，程序员不需要考虑底层的 utf-8 的细节。

go 也通过 rune 原生支持对 utf-8 的处理。

java 因为历史原因，其内部字符编码使用的是 utf-16，写程序时需要考虑编程语言内码与外码的转换。

另外各语言对解析字符串时的错误处理逻辑也有区别：

- Java/Go: 如果解码到无效的 bytes，将会返回一个特殊的字符 `\uFFFD`，打印出来是「�」
  - 因此如果遇到这个特殊 rune，通常意味着 string 的字符编码有错误
- Python: 会抛出异常，需要程序员主动处理异常

### 栈溢出与递归调用

许多语言都使用了一个固定大小的数据结构来实现调用栈，空间大小通常为 64kb 到 2MB.

这会导致的一个问题是，如果递归调用地太深了，就会导致栈空间用尽，程序运行失败，这被称作
`stack overflow` 栈溢出。

使用 Python 编写递归程序时就比较容易触发栈溢出的报错。

但是 Go 使用了一个变长的栈结构，栈空间可以动态扩容，因此使用 Go 编写程序不会遇到栈溢出报错（但是要注
意性能、空间消耗等问题）。

### 遍历数组时的一些坑

- 不要在遍历数组的同时，添加或删除其中的元素！

### 函数返回值

返回值这一块，不同的语言有不同的标准，有的语言如 Java/C# 不允许多返回值，多返回值必须使用 Class 封
装。而 Python/Go/C 等语言，都是支持多返回值的。

不过这里讨论的是另一个问题：如果一个函数定义了返回值，那它是否一定要以 `return` 结束？

对于绝大多数场景，上述问题的回答都是「是，一定要 `return`」。但是也存在特殊情况：

- 如果明确知道某个函数永远不会 `return`，那就不要求以 `return` 结束。

有两种情况会导致一个函数永远不会 `return`: 存在无限循环、或者存在 `panic`/异常.

这个规则在几乎所有语言中都适用，本人测试过 Go/Rust/Java/C# 都是这样。

比如下面这个 Go 程序，就可以正常编译并运行：

```go
package main

import "fmt"

func test() (int, int) {
	fmt.Println("inside test...")
	panic("failed")  // 这个函数永远会在这一步失败

  // 定义了返回值类型 (int, int)，但是未 return
}

func main() {
	fmt.Println("start...")
	test()
}
```

如下是另外一个无限循环的 Go 程序，同样可以正常编译运行：

```go
package main

import (
	"fmt"
	"time"
)

func test() (int, int) {
	fmt.Println("inside test...")
	for {
		time.Sleep(time.Second)
	}
	// 上面是一个无限循环，程序永远不会走到这一步

	// 定义了返回值类型 (int, int)，但是未 return
}

func main() {
	fmt.Println("start...")
	test()
}
```

如果把上面 Go 程序的无限循环改为有限，编译就会报错。

Rust 无限循环版本：

```rust
use std::{thread, time};

fn main() {
    println!("start...");
    test();
}

fn test() -> &'static str  {
    println!("inside test...");
    loop {  // 神奇的是这里如果用 while true 仍然会报错，用 loop 就能正常编译
        thread::sleep(time::Duration::from_millis(1000));
    }
}
```

Rust Panic 版本：

```rust
use std::{thread, time};

fn main() {
    println!("start...");
    test();
}

fn test() -> &'static str  {
    println!("inside test...");
    panic!("failed")
}
```

再来个 Java 的例子：

```java
import java.util.concurrent.TimeUnit;

class HelloWorld {
    public static void main(String[] args) {
        System.out.println("start...");
        try {
            test();
        } catch (Exception e) {
        }finally {
        }
    }

    public static int test() throws InterruptedException {
        System.out.println("inside test...");
        while (true) { // 无限循环
            TimeUnit.SECONDS.sleep(1);
        }
    }
}
```

编译运行这个 Java 程序：

```shell
❯ javac test.java
❯ java HelloWorld
start...
inside test...

```

检查下它的 JVM 字节码：

```

❯ javap -c HelloWorld
Compiled from "test.java"
class HelloWorld {
  HelloWorld();
    Code:
       0: aload_0
       1: invokespecial #1                  // Method java/lang/Object."<init>":()V
       4: return

  public static void main(java.lang.String[]);
    Code:
       0: getstatic     #2                  // Field java/lang/System.out:Ljava/io/PrintStream;
       3: ldc           #3                  // String start...
       5: invokevirtual #4                  // Method java/io/PrintStream.println:(Ljava/lang/String;)V
       8: invokestatic  #5                  // Method test:()I
      11: pop
      12: goto          22
      15: astore_1
      16: goto          22
      19: astore_2
      20: aload_2
      21: athrow
      22: return
    Exception table:
       from    to  target type
           8    12    15   Class java/lang/Exception
           8    12    19   any

  public static int test() throws java.lang.InterruptedException;
    Code:
       0: getstatic     #2                  // Field java/lang/System.out:Ljava/io/PrintStream;
       3: ldc           #7                  // String inside test...
       5: invokevirtual #4                  // Method java/io/PrintStream.println:(Ljava/lang/String;)V
       8: getstatic     #8                  // Field java/util/concurrent/TimeUnit.SECONDS:Ljava/util/concurrent/TimeUnit;
      11: lconst_1
      12: invokevirtual #9                  // Method java/util/concurrent/TimeUnit.sleep:(J)V
      15: goto          8
}
```

总之，能看到在上面几个示例程序中，`test()` 函数的返回类型定义完全是个多余的，居然也能正常编译过，这
样去思考的话，就觉得设计有点不一致。

但是跟朋友聊了后，意识到这个多余的类型，在调用方看来可能是必要的，比如说调用方可能就是要取这个类型的
返回值，你不定义的话就编译失败了。

大概理解了，进一步来说，这个永不返回的函数，其返回值可以定义成一个特殊的类型—— Never（比如 kotlin 里
就有 nothing 类型）. Never 可以被转换成任何其他类型。
